'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function constant(name, obj) {
  return this.factory(name, [], () => obj, {
    lifecycle: 'application'
  });
}

const isString = obj => typeof obj === 'string';
const isFunction = obj => typeof obj === 'function';
const instantiate = (context, args) => {
  // eslint-disable-next-line new-parens
  return new (Function.prototype.bind.apply(context, args))();
};
const isNode = () => {
  let _process; // eslint-disable-line no-underscore-dangle


  try {
    // eslint-disable-next-line no-new-func
    _process = new Function('return process')();
  } catch (e) {// No process
  } // eslint-disable-next-line max-len


  return typeof _process === 'object' && _process.toString && _process.toString() === '[object process]';
}; // eslint-disable-next-line no-new-func

const doUnsafeRequire = new Function('require', 'target', 'return require.main.require(target)');
const unsafeRequire = target => {
  // eslint-disable-next-line no-eval
  return doUnsafeRequire(eval('require'), target);
};
const ensureArray = arr => {
  if (arr == null) {
    return [];
  }

  if (Array.isArray(arr)) {
    return arr;
  }

  return [arr];
};
const hasLength = arr => arr == null || arr.length > 0;

const validateArgs = (name, dependencies, fn) => {
  if (!isString(name)) {
    throw new Error(`Factories must be given a name, but was called with [${typeof name}]`);
  }

  if (!Array.isArray(dependencies)) {
    throw new Error(`Expected an array of dependencies, but was called with [${typeof dependencies}]`);
  }

  if (!isFunction(fn)) {
    throw new Error(`Factory ${name} must be a [Function]`);
  }
};
function factory(name, dependencies, fn, opts = {}) {
  validateArgs(name, dependencies, fn);

  if (!hasLength(dependencies)) {
    dependencies = null;
  } // eslint-disable-next-line max-len


  if ((opts.precedence || this.$$config.precedence) === 'passive' && this.$$factories[name] != null) {
    return;
  }

  const f = {
    fn,
    dependencies,
    lifecycle: opts.lifecycle || this.$$config.lifecycle
  };
  this.$$factories[name] = f;

  if (opts.alias) {
    ensureArray(opts.alias).forEach(alias => this.alias(alias, name));
  }
}

function service(name, dependencies, fn, opts = {}) {
  validateArgs(name, dependencies, fn);

  function factory(...args) {
    const context = {};

    if (opts.bindToInstance) {
      dependencies.forEach((key, i) => {
        context[key] = args[i];
      });
    }

    return instantiate(fn, [context, ...args]);
  }

  return this.factory(name, dependencies, factory, opts);
}

function alias(alias, name) {
  if (this.$$alias[alias] == null || this.$$config.precedence === 'active') {
    this.$$alias[alias] = name;
  }

  if (this.$$alias[name] == null || this.$$config.precedence === 'active') {
    this.$$alias[name] = alias;
  }
}

const GLOBAL_TYPE_PREFIX = 'type:global:';
const NAMED_PARAMS = '$namedParameters';
const VOID = 'undefined';

const getFromNodeModules = (jpex, target) => {
  // in order to stop webpack environments from including every possible
  // import source in the bundle, we have to stick all node require stuff
  // inside an eval setup
  if (!jpex.$$config.nodeModules || !isNode()) {
    return;
  }

  try {
    const value = unsafeRequire(target);
    jpex.constant(target, value);
    return jpex.$$factories[target];
  } catch (e) {
    var _e$message;

    if ((_e$message = e.message) != null && _e$message.includes != null && _e$message.includes(`Cannot find module '${target}'`)) {
      // not found in node modules, just continue
      return;
    }

    throw e;
  }
};

const getGlobalObject = () => {
  if (typeof global !== VOID) {
    // eslint-disable-next-line no-undef
    return global;
  }

  if (typeof globalThis !== VOID) {
    // eslint-disable-next-line no-undef
    return globalThis;
  }

  if (typeof window !== VOID) {
    return window;
  }

  return {};
};

const getGlobalProperty = name => {
  const global = getGlobalObject();

  if (global[name] !== void 0) {
    return global[name];
  } // we need to handle inferred types as well
  // this gets a little bit hacky...


  if (name.startsWith(GLOBAL_TYPE_PREFIX)) {
    // most global types will just be the name of the property in pascal case
    // i.e. window = Window / document = Document
    const len = GLOBAL_TYPE_PREFIX.length;
    const inferredName = name.charAt(len).toLowerCase() + name.substr(len + 1);
    return global[inferredName];
  }
};

const getFromGlobal = (jpex, name) => {
  if (!jpex.$$config.globals) {
    return;
  }

  const value = getGlobalProperty(name);

  if (value !== void 0) {
    jpex.constant(name, value);
    return jpex.$$factories[name];
  }
};

const validateArgs$1 = name => {
  if (!isString(name)) {
    throw new Error(`Name must be a string, but recevied ${typeof name}`);
  }
};

const getFromAlias = (jpex, alias) => {
  const name = jpex.$$alias[alias];

  if (name != null) {
    return jpex.$$factories[name];
  }
};

const getFromResolved = (jpex, name) => {
  return jpex.$$resolved[name];
};

const getFromRegistry = (jpex, name) => {
  return jpex.$$factories[name];
};

const getFactory = (jpex, name, opts = {}) => {
  var _opts$optional;

  validateArgs$1(name);
  const fns = [getFromResolved, getFromRegistry, getFromAlias, getFromGlobal, getFromNodeModules];

  while (fns.length) {
    const factory = fns.shift()(jpex, name);

    if (factory != null) {
      return factory;
    }
  }

  if ((_opts$optional = opts.optional) != null ? _opts$optional : jpex.$$config.optional) {
    return;
  }

  throw new Error(`Unable to find required dependency [${name}]`);
};
const cacheResult = (jpex, name, factory, value, namedParameters) => {
  switch (factory.lifecycle) {
    case 'application':
      factory.resolved = true;
      factory.value = value;
      break;

    case 'class':
      jpex.$$resolved[name] = {
        resolved: true,
        value
      };
      break;

    case 'none':
      break;

    case 'instance':
    default:
      // instance
      namedParameters[name] = value;
      break;
  }
}; // Ensure we're not stuck in a recursive loop

const checkStack = (jpex, name, stack) => {
  if (!hasLength(stack)) {
    // This is the first loop
    return false;
  }

  if (!stack.includes(name)) {
    // We've definitely not tried to resolve this one before
    return false;
  }

  if (stack[stack.length - 1] === name) {
    var _jpex$$$parent;

    // We've tried to resolve this one before, but...
    // if this factory has overridden a parent factory
    // we should assume it actually wants to resolve the parent
    const parent = (_jpex$$$parent = jpex.$$parent) == null ? void 0 : _jpex$$$parent.$$factories[name];

    if (parent != null) {
      return true;
    }
  }

  throw new Error(`Recursive loop for dependency ${name} encountered`);
};

const getNamedParameters = (namedParameters, opts = {}) => {
  if (namedParameters) {
    return namedParameters;
  }

  if (opts.with) {
    return { ...opts.with
    };
  }

  return {};
};

const resolveFactory = (jpex, name, factory, namedParameters, opts, stack) => {
  if (factory == null) {
    return;
  } // Check if it's already been resolved


  if (factory.resolved) {
    return factory.value;
  } // Work out dependencies


  let args = [];

  if (hasLength(factory.dependencies)) {
    // eslint-disable-next-line no-use-before-define
    args = resolveMany(jpex, factory, namedParameters, opts, [...stack, name]);
  } // Invoke the factory


  const value = factory.fn.apply(jpex, args); // Cache the result

  cacheResult(jpex, name, factory, value, namedParameters);
  return value;
};

const resolveOne = (jpex, name, initialParameters, opts, stack) => {
  const namedParameters = getNamedParameters(initialParameters, opts); // Check named parameters
  // if we have a named parameter for this dependency
  // we don't need to do any resolution, we can just return the value

  if (Object.hasOwnProperty.call(namedParameters, name)) {
    return namedParameters[name];
  } // Special keys


  if (name === NAMED_PARAMS || name === "type:jpex/NamedParameters") {
    return namedParameters;
  }

  if (checkStack(jpex, name, stack)) {
    // Yes we have tried to resolve this one before, but we could
    // actually just be resolving an inherited factory
    return resolveOne(jpex.$$parent, name, namedParameters, opts, []);
  } // Get the factory
  // This will either return the factory,
  // return null (meaning it's an optional dependency)
  // or throw an error


  const factory = getFactory(jpex, name, opts);
  return resolveFactory(jpex, name, factory, namedParameters, opts, stack);
};
const resolveMany = (jpex, definition, namedParameters, opts, stack) => {
  if (!hasLength(definition.dependencies)) {
    return [];
  }

  if (!stack) {
    stack = [];
  }

  const dependencies = ensureArray(definition.dependencies);
  const values = dependencies.map(dependency => {
    return resolveOne(jpex, dependency, namedParameters, opts, stack);
  });
  return values;
};

function resolve(name, opts) {
  return resolveOne(this, name, void 0, opts, []);
}
function resolveDependencies(definition, opts) {
  return resolveMany(this, definition, void 0, opts, []);
}
function isResolved(dependency) {
  var _this$$$factories$dep;

  if (!isString(dependency)) {
    return false;
  }

  if (this.$$resolved[dependency] != null) {
    return true;
  }

  return ((_this$$$factories$dep = this.$$factories[dependency]) == null ? void 0 : _this$$$factories$dep.resolved) === true;
}
function allResolved(dependencies) {
  return dependencies.every(isResolved.bind(this));
}

function encase(dependencies, fn) {
  const jpex = this;
  let result;

  const encased = function (...args) {
    /* eslint-disable no-invalid-this */
    if (result && allResolved.call(jpex, dependencies)) {
      return result.apply(this, args);
    }

    const deps = resolveDependencies.call(jpex, {
      dependencies
    });
    result = fn.apply(jpex, deps);
    return result.apply(this, args);
    /* eslint-enable */
  };

  encased.encased = fn;
  return encased;
}

function clearCache(...names) {
  names = ensureArray(names);

  for (const key in this.$$factories) {
    if (!hasLength(names) || names.includes(key)) {
      this.$$factories[key].resolved = false;
    }
  }

  for (const key in this.$$resolved) {
    if (!hasLength(names) || names.includes(key)) {
      delete this.$$resolved[key];
    }
  }
}

const defaultConfig = {
  lifecycle: 'class',
  precedence: 'active',
  globals: true,
  nodeModules: true,
  optional: false
};
function makeJpex({
  inherit = true,
  ...config
} = {}, parent) {
  const jpex = {
    $$parent: parent,
    $$config: { ...defaultConfig,
      ...(inherit ? parent == null ? void 0 : parent.$$config : {}),
      ...config
    },
    $$factories: parent && inherit ? Object.create(parent.$$factories) : {},
    $$resolved: {},
    $$alias: parent && inherit ? Object.create(parent.$$alias) : {},
    constant,
    factory,
    service,
    alias,
    resolve,
    encase,
    clearCache,

    extend(config) {
      return makeJpex(config, this);
    },

    resolveWith(name, namedParameters, opts) {
      return this.resolve(name, {
        with: namedParameters,
        ...opts
      });
    },

    raw(name) {
      return getFactory(this, name, {}).fn;
    },

    infer: () => ''
  };
  return jpex;
}

const jpex = makeJpex();

exports.default = jpex;
exports.jpex = jpex;
