'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function constant(name, obj) {
  return this.factory(name, [], function () {
    return obj;
  }, {
    lifecycle: 'application'
  });
}

var isString = function isString(obj) {
  return typeof obj === 'string';
};
var isFunction = function isFunction(obj) {
  return typeof obj === 'function';
};
var instantiate = function instantiate(context, args) {
  // eslint-disable-next-line new-parens
  return new (Function.prototype.bind.apply(context, args))();
};
var isNode = function isNode() {
  var _process; // eslint-disable-line no-underscore-dangle


  try {
    // eslint-disable-next-line no-new-func
    _process = new Function('return process')();
  } catch (e) {// No process
  } // eslint-disable-next-line max-len


  return typeof _process === 'object' && _process.toString && _process.toString() === '[object process]';
}; // eslint-disable-next-line no-new-func

var doUnsafeRequire = new Function('require', 'target', 'return require.main.require(target)');
var unsafeRequire = function unsafeRequire(target) {
  // eslint-disable-next-line no-eval
  return doUnsafeRequire(eval('require'), target);
};
var ensureArray = function ensureArray(arr) {
  if (arr == null) {
    return [];
  }

  if (Array.isArray(arr)) {
    return arr;
  }

  return [arr];
};
var hasLength = function hasLength(arr) {
  return arr == null || arr.length > 0;
};

var validateArgs = function validateArgs(name, dependencies, fn) {
  if (!isString(name)) {
    throw new Error("Factories must be given a name, but was called with [" + typeof name + "]");
  }

  if (!Array.isArray(dependencies)) {
    throw new Error("Expected an array of dependencies, but was called with [" + typeof dependencies + "]");
  }

  if (!isFunction(fn)) {
    throw new Error("Factory " + name + " must be a [Function]");
  }
};
function factory(name, dependencies, fn, opts) {
  var _this = this;

  if (opts === void 0) {
    opts = {};
  }

  validateArgs(name, dependencies, fn);

  if (!hasLength(dependencies)) {
    dependencies = null;
  } // eslint-disable-next-line max-len


  if ((opts.precedence || this.$$config.precedence) === 'passive' && this.$$factories[name] != null) {
    return;
  }

  var f = {
    fn: fn,
    dependencies: dependencies,
    lifecycle: opts.lifecycle || this.$$config.lifecycle
  };
  this.$$factories[name] = f;

  if (opts.alias) {
    ensureArray(opts.alias).forEach(function (alias) {
      return _this.alias(alias, name);
    });
  }
}

function service(name, dependencies, fn, opts) {
  if (opts === void 0) {
    opts = {};
  }

  validateArgs(name, dependencies, fn);

  function factory() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var context = {};

    if (opts.bindToInstance) {
      dependencies.forEach(function (key, i) {
        context[key] = args[i];
      });
    }

    return instantiate(fn, [context].concat(args));
  }

  return this.factory(name, dependencies, factory, opts);
}

function alias(alias, name) {
  if (this.$$alias[alias] == null || this.$$config.precedence === 'active') {
    this.$$alias[alias] = name;
  }

  if (this.$$alias[name] == null || this.$$config.precedence === 'active') {
    this.$$alias[name] = alias;
  }
}

var GLOBAL_TYPE_PREFIX = 'type:global:';
var NAMED_PARAMS = '$namedParameters';
var VOID = 'undefined';

var getFromNodeModules = function getFromNodeModules(jpex, target) {
  // in order to stop webpack environments from including every possible
  // import source in the bundle, we have to stick all node require stuff
  // inside an eval setup
  if (!jpex.$$config.nodeModules || !isNode()) {
    return;
  }

  try {
    var value = unsafeRequire(target);
    jpex.constant(target, value);
    return jpex.$$factories[target];
  } catch (e) {
    var _e$message;

    if ((_e$message = e.message) != null && _e$message.includes != null && _e$message.includes("Cannot find module '" + target + "'")) {
      // not found in node modules, just continue
      return;
    }

    throw e;
  }
};

var getGlobalObject = function getGlobalObject() {
  if (typeof global !== VOID) {
    // eslint-disable-next-line no-undef
    return global;
  }

  if (typeof globalThis !== VOID) {
    // eslint-disable-next-line no-undef
    return globalThis;
  }

  if (typeof window !== VOID) {
    return window;
  }

  return {};
};

var getGlobalProperty = function getGlobalProperty(name) {
  var global = getGlobalObject();

  if (global[name] !== void 0) {
    return global[name];
  } // we need to handle inferred types as well
  // this gets a little bit hacky...


  if (name.startsWith(GLOBAL_TYPE_PREFIX)) {
    // most global types will just be the name of the property in pascal case
    // i.e. window = Window / document = Document
    var len = GLOBAL_TYPE_PREFIX.length;
    var inferredName = name.charAt(len).toLowerCase() + name.substr(len + 1);
    return global[inferredName];
  }
};

var getFromGlobal = function getFromGlobal(jpex, name) {
  if (!jpex.$$config.globals) {
    return;
  }

  var value = getGlobalProperty(name);

  if (value !== void 0) {
    jpex.constant(name, value);
    return jpex.$$factories[name];
  }
};

var validateArgs$1 = function validateArgs(name) {
  if (!isString(name)) {
    throw new Error("Name must be a string, but recevied " + typeof name);
  }
};

var getFromAlias = function getFromAlias(jpex, alias) {
  var name = jpex.$$alias[alias];

  if (name != null) {
    return jpex.$$factories[name];
  }
};

var getFromResolved = function getFromResolved(jpex, name) {
  return jpex.$$resolved[name];
};

var getFromRegistry = function getFromRegistry(jpex, name) {
  return jpex.$$factories[name];
};

var getFactory = function getFactory(jpex, name, opts) {
  var _opts$optional;

  if (opts === void 0) {
    opts = {};
  }

  validateArgs$1(name);
  var fns = [getFromResolved, getFromRegistry, getFromAlias, getFromGlobal, getFromNodeModules];

  while (fns.length) {
    var factory = fns.shift()(jpex, name);

    if (factory != null) {
      return factory;
    }
  }

  if ((_opts$optional = opts.optional) != null ? _opts$optional : jpex.$$config.optional) {
    return;
  }

  throw new Error("Unable to find required dependency [" + name + "]");
};
var cacheResult = function cacheResult(jpex, name, factory, value, namedParameters) {
  switch (factory.lifecycle) {
    case 'application':
      factory.resolved = true;
      factory.value = value;
      break;

    case 'class':
      jpex.$$resolved[name] = {
        resolved: true,
        value: value
      };
      break;

    case 'none':
      break;

    case 'instance':
    default:
      // instance
      namedParameters[name] = value;
      break;
  }
}; // Ensure we're not stuck in a recursive loop

var checkStack = function checkStack(jpex, name, stack) {
  if (!hasLength(stack)) {
    // This is the first loop
    return false;
  }

  if (!stack.includes(name)) {
    // We've definitely not tried to resolve this one before
    return false;
  }

  if (stack[stack.length - 1] === name) {
    var _jpex$$$parent;

    // We've tried to resolve this one before, but...
    // if this factory has overridden a parent factory
    // we should assume it actually wants to resolve the parent
    var parent = (_jpex$$$parent = jpex.$$parent) == null ? void 0 : _jpex$$$parent.$$factories[name];

    if (parent != null) {
      return true;
    }
  }

  throw new Error("Recursive loop for dependency " + name + " encountered");
};

var getNamedParameters = function getNamedParameters(namedParameters, opts) {
  if (opts === void 0) {
    opts = {};
  }

  if (namedParameters) {
    return namedParameters;
  }

  if (opts.with) {
    return _extends({}, opts.with);
  }

  return {};
};

var resolveFactory = function resolveFactory(jpex, name, factory, namedParameters, opts, stack) {
  if (factory == null) {
    return;
  } // Check if it's already been resolved


  if (factory.resolved) {
    return factory.value;
  } // Work out dependencies


  var args = [];

  if (hasLength(factory.dependencies)) {
    // eslint-disable-next-line no-use-before-define
    args = resolveMany(jpex, factory, namedParameters, opts, [].concat(stack, [name]));
  } // Invoke the factory


  var value = factory.fn.apply(jpex, args); // Cache the result

  cacheResult(jpex, name, factory, value, namedParameters);
  return value;
};

var resolveOne = function resolveOne(jpex, name, initialParameters, opts, stack) {
  var namedParameters = getNamedParameters(initialParameters, opts); // Check named parameters
  // if we have a named parameter for this dependency
  // we don't need to do any resolution, we can just return the value

  if (Object.hasOwnProperty.call(namedParameters, name)) {
    return namedParameters[name];
  } // Special keys


  if (name === NAMED_PARAMS || name === "type:jpex/NamedParameters") {
    return namedParameters;
  }

  if (checkStack(jpex, name, stack)) {
    // Yes we have tried to resolve this one before, but we could
    // actually just be resolving an inherited factory
    return resolveOne(jpex.$$parent, name, namedParameters, opts, []);
  } // Get the factory
  // This will either return the factory,
  // return null (meaning it's an optional dependency)
  // or throw an error


  var factory = getFactory(jpex, name, opts);
  return resolveFactory(jpex, name, factory, namedParameters, opts, stack);
};
var resolveMany = function resolveMany(jpex, definition, namedParameters, opts, stack) {
  if (!hasLength(definition.dependencies)) {
    return [];
  }

  if (!stack) {
    stack = [];
  }

  var dependencies = ensureArray(definition.dependencies);
  var values = dependencies.map(function (dependency) {
    return resolveOne(jpex, dependency, namedParameters, opts, stack);
  });
  return values;
};

function resolve(name, opts) {
  return resolveOne(this, name, void 0, opts, []);
}
function resolveDependencies(definition, opts) {
  return resolveMany(this, definition, void 0, opts, []);
}
function isResolved(dependency) {
  var _this$$$factories$dep;

  if (!isString(dependency)) {
    return false;
  }

  if (this.$$resolved[dependency] != null) {
    return true;
  }

  return ((_this$$$factories$dep = this.$$factories[dependency]) == null ? void 0 : _this$$$factories$dep.resolved) === true;
}
function allResolved(dependencies) {
  return dependencies.every(isResolved.bind(this));
}

function encase(dependencies, fn) {
  var jpex = this;
  var result;

  var encased = function encased() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    /* eslint-disable no-invalid-this */
    if (result && allResolved.call(jpex, dependencies)) {
      return result.apply(this, args);
    }

    var deps = resolveDependencies.call(jpex, {
      dependencies: dependencies
    });
    result = fn.apply(jpex, deps);
    return result.apply(this, args);
    /* eslint-enable */
  };

  encased.encased = fn;
  return encased;
}

function clearCache() {
  for (var _len = arguments.length, names = new Array(_len), _key = 0; _key < _len; _key++) {
    names[_key] = arguments[_key];
  }

  names = ensureArray(names);

  for (var key in this.$$factories) {
    if (!hasLength(names) || names.includes(key)) {
      this.$$factories[key].resolved = false;
    }
  }

  for (var _key2 in this.$$resolved) {
    if (!hasLength(names) || names.includes(_key2)) {
      delete this.$$resolved[_key2];
    }
  }
}

var defaultConfig = {
  lifecycle: 'class',
  precedence: 'active',
  globals: true,
  nodeModules: true,
  optional: false
};
function makeJpex(_temp, parent) {
  var _ref = _temp === void 0 ? {} : _temp,
      _ref$inherit = _ref.inherit,
      inherit = _ref$inherit === void 0 ? true : _ref$inherit,
      config = _objectWithoutPropertiesLoose(_ref, ["inherit"]);

  var jpex = {
    $$parent: parent,
    $$config: _extends({}, defaultConfig, inherit ? parent == null ? void 0 : parent.$$config : {}, config),
    $$factories: parent && inherit ? Object.create(parent.$$factories) : {},
    $$resolved: {},
    $$alias: parent && inherit ? Object.create(parent.$$alias) : {},
    constant: constant,
    factory: factory,
    service: service,
    alias: alias,
    resolve: resolve,
    encase: encase,
    clearCache: clearCache,
    extend: function extend(config) {
      return makeJpex(config, this);
    },
    resolveWith: function resolveWith(name, namedParameters, opts) {
      return this.resolve(name, _extends({
        with: namedParameters
      }, opts));
    },
    raw: function raw(name) {
      return getFactory(this, name, {}).fn;
    },
    infer: function infer() {
      return '';
    }
  };
  return jpex;
}

var jpex = makeJpex();

exports.default = jpex;
exports.jpex = jpex;
